(************************************
 * definition of specification
 ************************************)
Require Import EqNat.
Require Export tm.
Require Export natutil.
Require Export map.

(** predicate *)
Definition pred := (code -> state -> index -> Prop)%type.

(** action *)
Definition action := (code -> state -> state -> index -> Prop)%type.

(** monotonic predicate *)
Definition monop (p : pred) :=
  forall (c : code) (s : state) (i j : index),
    p c s i -> i > j -> p c s j.

Definition monog (g : action) :=
  forall (c : code) (s s' : state) (i j : index),
    g c s s' i -> i > j -> g c s s' j.

Definition pand (p p' : pred) : pred :=
  fun (c : code) (s : state) (i : index) =>
    p c s i /\ p' c s i.

Definition por (p p' : pred) : pred :=
  fun (c : code) (s : state) (i : index) =>
    p c s i \/ p' c s i.

Definition pimp (p p' : pred) : Prop :=
  forall (c : code) (s : state) (i : index),
    p c s i -> p' c s i.

Definition psep (p p' : pred) : pred :=
  fun (c : code) (s : state) (i : index) =>
    exists s1 : state, exists s2 : state,
      StateMod.disjUnion s1 s2 s /\ p c s1 i /\ p' c s2 i.

Definition gand (g g' : action) : action :=
  fun (c : code) (s s' : state) (i : index) =>
    g c s s' i /\ g' c s s' i.

Definition gor (g g' : action) : action :=
  fun (c : code) (s s' : state) (i : index) =>
    g c s s' i \/ g' c s s' i.

Definition gimp (g g' : action) : Prop :=
  forall (c : code) (s s' : state) (i : index),
    g c s s' i -> g' c s s' i.

Definition pgtog (p : pred) (g : action) : action :=
  fun (c : code) (s s' : state) (i : index) =>
    p c s i /\ g c s s' i.

Definition pgtop (p : pred) (g : action) : pred :=
  fun (c : code) (s : state) (i : index) =>
    exists s' : state, p c s' i /\ g c s' s i.

Definition gpsep (g : action ) (p : pred) : action :=
  fun (c : code) (s s' : state) (i : index) =>
    exists s1 : state, exists s2 : state, exists s1' : state,
      StateMod.disjUnion s1 s2 s /\ StateMod.disjUnion s1' s2 s' /\
      g c s1 s1' i /\ p c s2 i.

Definition gsep (g g' : action) : action :=
  fun (c : code) (s s' : state) (i : index) =>
    exists s1 : state, exists s2 : state, exists s1' : state, exists s2' : state,
      StateMod.disjUnion s1 s2 s /\ StateMod.disjUnion s1' s2' s' /\
      g c s1 s1' i /\ g' c s2 s2' i.

Definition glink (g g' : action) : action :=
  fun (c : code) (s s' : state) (i : index) =>
    exists s'' : state, 
      g c s s'' i /\ g' c s'' s' i.

(** special specifications *)
Definition rptox (r : reg) : pred :=
  fun (c : code) (s : state) (i : index) =>
    exists v : val, StateMod.sgltonM s (regR r) v.

Definition rptoxx (r : reg) : action :=
  fun (c : code) (s s': state) (i : index) =>
    exists v : val, exists v' : val, 
      StateMod.sgltonM s (regR r) v /\ StateMod.sgltonM s' (regR r) v'.
  
Definition rgtz (r : reg) : pred :=
  fun (c : code) (s : state) (i : index) =>
    exists v : val, StateMod.maps_to s (regR r) v /\ v > 0.

Definition rlez (r : reg) : pred :=
  fun (c : code) (s : state) (i : index) =>
    exists v : val, StateMod.maps_to s (regR r) v /\ v <= 0.

Definition gid : action :=
  fun (c : code) (s s' : state) (i : index) => s = s'.

Definition spec := (pred * action)%type.


(** lemmas *)
Lemma monop_rptox : 
  forall (r : reg), monop (rptox r).
Proof.
unfold monop, rptox.
intros.
assumption.
Qed.

Lemma monop_psep :
  forall (p p' : pred),
    monop p -> monop p' -> monop (psep p p').
Proof.
unfold monop.
intros.
destruct H1 as [s1 [s2 [I1 [I2 I3]]]].
exists s1; exists s2.
split.
assumption.
split.
apply H with i; assumption.
apply H0 with i; assumption.
Qed.